

class Intent:

    NO_ACTION = 'no_action'

    NO_RESULT = 'no_result'

    def __init__(self, presenter_class: type):
        self.__presenter_data = {}
        self.__presenter_action = self.NO_ACTION
        self.__presenter_class = presenter_class
        self.__use_new_window = False
        self.__use_modal = False

    def get_presenter_class(self) -> type:
        return self.__presenter_class

    def get_data(self) -> dict:
        return self.__presenter_data

    def set_data(self, data: dict):
        self.__presenter_data = data

    def get_action(self) -> str:
        return self.__presenter_action

    def set_action(self, action: str):
        self.__presenter_action = action

    def use_new_window(self, use_new_window: bool = False):
        self.__use_new_window = use_new_window

    def is_using_new_window(self) -> bool:
        return self.__use_new_window

    def use_modal(self, modal: bool = False):
        self.__use_modal = modal

    def is_using_modal(self) -> bool:
        return self.__use_modal



from PyQt5.QtWidgets import QApplication
from easy_mvp.abstract_presenter import AbstractPresenter
from easy_mvp.exception import NoGlobalDataWithKeyException
from easy_mvp.intent import Intent
from easy_mvp.window import WindowHandler


class ApplicationManager:

    def __init__(self):
        self.__window_stack = []
        self.__app = QApplication([])
        self.__global_data = {}

    def add_new_window(self, intent: Intent, parent_window: WindowHandler, calling_presenter: AbstractPresenter):
        window = WindowHandler(self, parent_window)
        window.add_presenter(intent, calling_presenter)
        self.__window_stack.append(window)
        window.show()

    def remove_window(self, window: WindowHandler):
        self.__window_stack.remove(window)

    def execute_app(self, initial_intent: Intent):
        window = WindowHandler(self)
        window.add_presenter(initial_intent)

        self.__window_stack.append(window)
        window.show()
        self.__app.exec()

    def exit(self, code: int = 0):
        self.__app.exit(code)

    def set_global_data(self, key: str, data):
        self.__global_data[key] = data

    def get_global_data(self, key: str):
        if key not in self.__global_data:
            raise NoGlobalDataWithKeyException(key)
        return self.__global_data[key]

    def has_global_data(self, key: str) -> bool:
        return key in self.__global_data



from easy_mvp.intent import Intent


class AbstractPresenter:

    def __init__(self, intent: Intent, window, app_manager):
        self.__intent = intent
        self.__window_handler = window
        self.__app_manager = app_manager
        self.__view = None
        self._on_initialize()

    def get_view(self):
        return self.__view

    def _set_view(self, view):
        self.__view = view

    def _open_other_presenter(self, intent: Intent):
        if intent.is_using_new_window():
            self.__app_manager.add_new_window(intent,
                                              parent_window=self.__window_handler,
                                              calling_presenter=self)
        else:
            self.__window_handler.add_presenter(intent, self)

    def _close_this_presenter(self):
        self.__window_handler.pop_presenter(self)

    def _close_this_presenter_with_result(self, result_data: dict, result: str = Intent.NO_RESULT):
        self.__window_handler.pop_presenter_with_result(self.__intent,
                                                        calling_presenter=self,
                                                        result_data=result_data,
                                                        result=result)

    def _on_initialize(self):
        pass

    def on_view_shown(self):
        pass

    def on_view_covered(self):
        pass

    def on_view_discovered(self):
        pass

    def on_view_discovered_with_result(self, action: str, result_data: dict, result: str):
        pass

    def on_closing_presenter(self):
        pass

    def _get_intent_data(self) -> dict:
        return self.__intent.get_data()

    def _get_intent_action(self) -> str:
        return self.__intent.get_action()

    def exit_app(self, code: int = 0):
        self.__window_handler.exit_app(code)

    def set_global_data(self, key: str, data):
        self.__app_manager.set_global_data(key, data)

    def get_global_data(self, key: str):
        return self.__app_manager.get_global_data(key)

    def has_global_data(self, key: str) -> bool:
        return self.__app_manager.has_global_data(key)

    def get_default_window_title(self) -> str:
        # This method will be called before the presenter receives the
        # on_view_show, on_view_discovered and on_view_discovered_with_result
        # calls.
        return 'No Title, reimplement get_default_window_title'

    def _set_window_title(self, window_title: str):
        self.__window_handler.set_window_title(window_title)

    def on_window_closing(self):
        pass





class BelowPresenterDoingCommandsException(Exception):

    def __init__(self):
        super().__init__('A presenter which is below in the stack is doing commands.\n' +
                         'For safety and clarity it is not possible for a below presenter\n' +
                         'to do commands. The top presenter is the only one who can do\n' +
                         'commands. In this way the stack model is always reached.')


class NoBelowPresenterToBeNotifiedWithResultException(Exception):

    def __init__(self):
        super().__init__('A presenter can not be closed with result if there is not a below\n' +
                         'presenter on the same window. If there is not below presenter then\n' +
                         'it is necessary to the presenter\'s window to have a parent window.')


class NoGlobalDataWithKeyException(Exception):

    def __init__(self, key: str):
        super().__init__('No global data with key {}.'.format(key))



from PyQt5.QtGui import QCloseEvent
from PyQt5.QtWidgets import QStackedWidget
from PyQt5.QtCore import Qt
from easy_mvp.abstract_presenter import AbstractPresenter
from easy_mvp.exception import BelowPresenterDoingCommandsException, NoBelowPresenterToBeNotifiedWithResultException
from easy_mvp.intent import Intent


class StackWindow(QStackedWidget):

    def __init__(self, window_handler):
        super().__init__()
        self.__window_handler = window_handler

    def closeEvent(self, event: QCloseEvent):
        super().closeEvent(event)


class WindowHandler:

    def __init__(self, application_manager, parent_window=None):
        self.__presenter_stack = []
        self.__parent_window = None
        self.__link_to_parent_window(parent_window)
        self.__stacked_widget = StackWindow(self)
        self.__app_manager = application_manager
        self.__child_windows = []

    def __link_to_parent_window(self, parent_window):
        if parent_window is not None:
            parent_window.add_child_window(self)
        self.__parent_window = parent_window

    def get_base_widget(self) -> QStackedWidget:
        return self.__stacked_widget

    def add_presenter(self, intent: Intent, calling_presenter: AbstractPresenter = None):

        if self.presenter_count() == 0:
            self.__set_modal(intent.is_using_modal())

        self.__check_is_top_presenter(calling_presenter)
        self.__notify_presenter_on_view_covered(calling_presenter)

        self.__add_presenter_and_its_view(intent)
        self.update_window_title()
        self.__notify_top_presenter_on_view_shown()

    def presenter_count(self) -> int:
        return len(self.__presenter_stack)

    def __set_modal(self, modal: bool):
        if modal:
            self.__stacked_widget.setWindowModality(Qt.WindowModality.ApplicationModal)
        else:
            self.__stacked_widget.setWindowModality(Qt.WindowModality.NonModal)

    def __check_is_top_presenter(self, presenter: AbstractPresenter):
        if self.presenter_count() >= 1:
            top_presenter = self.get_top_presenter()
            if top_presenter is not presenter:
                raise BelowPresenterDoingCommandsException()

    @staticmethod
    def __notify_presenter_on_view_covered(self, calling_presenter: AbstractPresenter = None):
        if calling_presenter is not None:
            calling_presenter.on_view_covered()

    def __add_presenter_and_its_view(self, intent: Intent):
        presenter_class = intent.get_presenter_class()
        new_presenter = presenter_class(intent, self, self.__app_manager)
        self.__presenter_stack.append(new_presenter)
        self.__stacked_widget.addWidget(new_presenter.get_view())
        self.__stacked_widget.setCurrentWidget(new_presenter.get_view())

    def update_window_title(self):
        top_presenter_title = self.get_top_presenter().get_default_window_title()
        self.__stacked_widget.setWindowTitle(top_presenter_title)

    def __notify_top_presenter_on_view_shown(self):
        self.get_top_presenter().on_view_shown()

    def get_top_presenter(self) -> AbstractPresenter:
        return self.__presenter_stack[-1]

    def pop_presenter(self, calling_presenter: AbstractPresenter) -> AbstractPresenter:
        self.__check_is_top_presenter(calling_presenter)

        top_presenter = self.__pop_presenter_and_its_view()
        top_presenter.on_closing_presenter()
        top_presenter.get_view().deleteLater()

        was_window_closed = self.__close_window_if_no_presenter_remains()
        if was_window_closed and self.has_parent_window():
            self.__parent_window.update_window_title()
            presenter_of_parent_window = self.__parent_window.get_top_presenter()
            presenter_of_parent_window.on_view_discovered()
        elif self.presenter_count() >= 1:
            below_presenter = self.get_top_presenter()
            self.update_window_title()
            below_presenter.on_view_discovered()
        else:
            self.__app_manager.exit()

        return top_presenter

    def __pop_presenter_and_its_view(self) -> AbstractPresenter:
        top_presenter = self.__presenter_stack.pop(-1)
        self.__stacked_widget.removeWidget(top_presenter.get_view())
        return top_presenter

    def __notify_presenter_on_view_discovered(self):
        if len(self.__presenter_stack) >= 1:
            below_presenter = self.get_top_presenter()
            below_presenter.on_view_discovered()

    def __close_window_if_no_presenter_remains(self) -> bool:
        if len(self.__presenter_stack) == 0:
            self.close_window()
            return True
        return False

    def close_window(self):
        self.__stacked_widget.close()

    def remove_from_app_manager(self):
        self.notify_presenters_on_window_closing()
        self.close_all_child_windows()
        
        self.remove_from_app_manager()
        self.remove_from_parent_window_if_possible()
        
        self.__stacked_widget.close()
        self.__app_manager.remove_window(self)
    
    def remove_from_parent_window_if_possible(self):
        if self.__parent_window is not None:
            self.__parent_window.remove_child_window(self)

    def notify_presenters_on_window_closing(self):
        last_presenter_stack_index = self.presenter_count() - 1

        # iterate over __presenter_stack reversely
        for presenter_index in range(last_presenter_stack_index, -1, -1):
            a_presenter: AbstractPresenter = self.__presenter_stack[presenter_index]
            a_presenter.on_window_closing()

    def pop_presenter_with_result(self, intent: Intent, calling_presenter: AbstractPresenter, result_data: dict, result: str):
        self.__check_is_top_presenter(calling_presenter)
        self.__check_there_is_below_presenter_to_be_notified_with_result()

        top_presenter = self.__pop_presenter_and_its_view()
        top_presenter.on_closing_presenter()
        top_presenter.get_view().deleteLater()

        was_window_closed = self.__close_window_if_no_presenter_remains()
        if was_window_closed:
            self.__parent_window.update_window_title()
            self.__notify_presenter_on_discovered_with_result_on_parent_window(intent, result_data, result)
        else:
            self.update_window_title()
            self.__notify_below_presenter_on_discovered_with_result(intent, result_data, result)

    def __check_there_is_below_presenter_to_be_notified_with_result(self):
        if self.presenter_count() == 1 and not self.has_parent_window():
            raise NoBelowPresenterToBeNotifiedWithResultException()

    def has_parent_window(self) -> bool:
        return self.__parent_window is not None

    def __notify_presenter_on_discovered_with_result_on_parent_window(self, intent: Intent, result_data: dict, result: str):
        parent_window_presenter = self.__parent_window.get_top_presenter()
        parent_window_presenter.on_view_discovered_with_result(intent.get_action(), result_data, result)

    def __notify_below_presenter_on_discovered_with_result(self, intent: Intent, result_data: dict, result: str):
        below_presenter = self.get_top_presenter()
        below_presenter.on_view_discovered_with_result(intent.get_action(), result_data, result)

    def show(self):
        self.__stacked_widget.show()

    def exit_app(self, code: int = 0):
        self.__app_manager.exit(code)

    def close_all_child_windows(self):
        while len(self.__child_windows) > 0:
            child_window = self.__child_windows[0]
            child_window.close_window()

    def add_child_window(self, child_window):
        self.__child_windows.append(child_window)

    def remove_child_window(self, child_window):
        self.__child_windows.remove(child_window)

    def set_window_title(self, window_title: str):
        self.__stacked_widget.setWindowTitle(window_title)
